@@ -7,20 +7,25 @@
 package azidentity
 
 import (
+	"bytes"
 	"context"
 	"fmt"
 	"net/http"
 	"net/url"
 	"os"
+	"os/exec"
 	"path/filepath"
+	"runtime"
 	"strings"
 	"testing"
 	"time"
 
 	"github.com/Azure/azure-sdk-for-go/sdk/azcore"
 	"github.com/Azure/azure-sdk-for-go/sdk/azcore/policy"
+	azruntime "github.com/Azure/azure-sdk-for-go/sdk/azcore/runtime"
 	"github.com/Azure/azure-sdk-for-go/sdk/internal/mock"
 	"github.com/Azure/azure-sdk-for-go/sdk/internal/recording"
+	"github.com/stretchr/testify/require"
 )
 
 const (
@@ -28,15 +33,12 @@ const (
 	expiresOnNonStringIntResp = `{"access_token": "new_token", "refresh_token": "", "expires_in": "", "expires_on": 1560974028, "not_before": "1560970130", "resource": "https://vault.azure.net", "token_type": "Bearer"}`
 )
 
-// TODO: replace with 1.17's T.Setenv
-func clearEnvVars(envVars ...string) {
-	for _, ev := range envVars {
-		_ = os.Unsetenv(ev)
-	}
-}
-
 func TestManagedIdentityCredential_AzureArc(t *testing.T) {
-	file, err := os.Create(filepath.Join(t.TempDir(), "arc.key"))
+	d := t.TempDir()
+	before := arcKeyDirectory
+	arcKeyDirectory = func() (string, error) { return d, nil }
+	defer func() { arcKeyDirectory = before }()
+	file, err := os.Create(filepath.Join(d, "arc.key"))
 	if err != nil {
 		t.Fatal(err)
 	}
@@ -154,6 +156,107 @@ func TestManagedIdentityCredential_AzureArcErrors(t *testing.T) {
 			t.Fatal("expected an error")
 		}
 	})
+	t.Run("key too large", func(t *testing.T) {
+		d := t.TempDir()
+		f := filepath.Join(d, "test.key")
+		err := os.WriteFile(f, bytes.Repeat([]byte("."), 4097), 0600)
+		require.NoError(t, err)
+		before := arcKeyDirectory
+		arcKeyDirectory = func() (string, error) { return d, nil }
+		defer func() { arcKeyDirectory = before }()
+		srv, close := mock.NewServer(mock.WithTransformAllRequestsToTestServerUrl())
+		defer close()
+		srv.AppendResponse(
+			mock.WithHeader("WWW-Authenticate", "Basic realm="+f),
+			mock.WithStatusCode(http.StatusUnauthorized),
+		)
+		cred, err := NewManagedIdentityCredential(&ManagedIdentityCredentialOptions{ClientOptions: azcore.ClientOptions{Transport: srv}})
+		require.NoError(t, err)
+		_, err = cred.GetToken(ctx, testTRO)
+		require.ErrorContains(t, err, "too large")
+	})
+	t.Run("unexpected file paths", func(t *testing.T) {
+		d, err := arcKeyDirectory()
+		if err != nil {
+			// test is running on an unsupported OS e.g. darwin
+			t.Skip(err)
+		}
+		srv, close := mock.NewServer(mock.WithTransformAllRequestsToTestServerUrl())
+		defer close()
+		srv.AppendResponse(
+			// unexpected directory
+			mock.WithHeader("WWW-Authenticate", "Basic realm="+filepath.Join("foo", "bar.key")),
+			mock.WithStatusCode(http.StatusUnauthorized),
+		)
+		o := ManagedIdentityCredentialOptions{ClientOptions: azcore.ClientOptions{Transport: srv}}
+		cred, err := NewManagedIdentityCredential(&o)
+		require.NoError(t, err)
+		_, err = cred.GetToken(ctx, testTRO)
+		require.ErrorContains(t, err, "unexpected file path")
+
+		srv.AppendResponse(
+			// unexpected extension
+			mock.WithHeader("WWW-Authenticate", "Basic realm="+filepath.Join(d, "foo")),
+			mock.WithStatusCode(http.StatusUnauthorized),
+		)
+		cred, err = NewManagedIdentityCredential(&o)
+		require.NoError(t, err)
+		_, err = cred.GetToken(ctx, testTRO)
+		require.ErrorContains(t, err, "unexpected file path")
+	})
+	if runtime.GOOS == "windows" {
+		t.Run("ProgramData not set", func(t *testing.T) {
+			t.Setenv("ProgramData", "")
+			srv, close := mock.NewServer(mock.WithTransformAllRequestsToTestServerUrl())
+			defer close()
+			srv.AppendResponse(
+				mock.WithHeader("WWW-Authenticate", "Basic realm=foo"),
+				mock.WithStatusCode(http.StatusUnauthorized),
+			)
+			cred, err := NewManagedIdentityCredential(&ManagedIdentityCredentialOptions{ClientOptions: azcore.ClientOptions{Transport: srv}})
+			require.NoError(t, err)
+			_, err = cred.GetToken(ctx, testTRO)
+			require.ErrorContains(t, err, "ProgramData")
+		})
+	}
+}
+
+func TestManagedIdentityCredential_AzureContainerInstanceLive(t *testing.T) {
+	// This test triggers the managed identity test app deployed to an Azure Container Instance.
+	// See the bicep file and test resources scripts for details.
+	// It triggers the app with az because the test subscription prohibits opening ports to the internet.
+	name := os.Getenv("AZIDENTITY_ACI_NAME")
+	rg := os.Getenv("AZIDENTITY_RESOURCE_GROUP")
+	if name == "" || rg == "" {
+		t.Skip("set AZIDENTITY_ACI_NAME and AZIDENTITY_RESOURCE_GROUP to run this test")
+	}
+	ctx, cancel := context.WithTimeout(context.Background(), time.Minute)
+	defer cancel()
+	command := fmt.Sprintf("az container exec -g %s -n %s --exec-command 'wget -qO- localhost'", rg, name)
+	// using "script" as a workaround for "az container exec" requiring a tty
+	// https://github.com/Azure/azure-cli/issues/17530
+	cmd := exec.CommandContext(ctx, "script", "-q", "-O", "/dev/null", "-c", command)
+	b, err := cmd.CombinedOutput()
+	s := string(b)
+	require.NoError(t, err, s)
+	require.Equal(t, "test passed", s)
+}
+
+func TestManagedIdentityCredential_AzureFunctionsLive(t *testing.T) {
+	// This test triggers the managed identity test app deployed to Azure Functions.
+	// See the bicep file and test resources scripts for details.
+	fn := os.Getenv("AZIDENTITY_FUNCTION_NAME")
+	if fn == "" {
+		t.Skip("set AZIDENTITY_FUNCTION_NAME to run this test")
+	}
+	url := fmt.Sprintf("https://%s.azurewebsites.net/api/HttpTrigger", fn)
+	res, err := http.Get(url)
+	require.NoError(t, err)
+	if res.StatusCode != http.StatusOK {
+		b, err := azruntime.Payload(res)
+		require.NoError(t, err)
+		t.Fatal("test application returned an error: " + string(b))
+	}
 }
 
 func TestManagedIdentityCredential_AzureMLLive(t *testing.T) {
@@ -300,8 +403,8 @@ func TestManagedIdentityCredential_GetTokenIMDS400(t *testing.T) {
 	// cred should return credentialUnavailableError when IMDS responds 400 to a token request
 	for i := 0; i < 3; i++ {
 		_, err = cred.GetToken(context.Background(), testTRO)
-		if _, ok := err.(*credentialUnavailableError); !ok {
-			t.Fatalf("expected credentialUnavailableError, received %T", err)
+		if _, ok := err.(credentialUnavailable); !ok {
+			t.Fatalf("expected credentialUnavailable, received %T", err)
 		}
 	}
 }
@@ -472,61 +575,56 @@ func TestManagedIdentityCredential_CreateAccessTokenExpiresOnFail(t *testing.T)
 }
 
 func TestManagedIdentityCredential_IMDSLive(t *testing.T) {
-	switch recording.GetRecordMode() {
-	case recording.LiveMode:
-		t.Skip("this test doesn't run in live mode because it can't pass in CI")
-	case recording.RecordingMode:
-		// record iff either managed identity environment variable is set, because
-		// otherwise there's no reason to believe the test is running on a VM
-		if len(liveManagedIdentity.clientID)+len(liveManagedIdentity.resourceID) == 0 {
-			t.Skip("neither MANAGED_IDENTITY_CLIENT_ID nor MANAGED_IDENTITY_RESOURCE_ID is set")
-		}
+	if recording.GetRecordMode() != recording.PlaybackMode && !liveManagedIdentity.imds {
+		t.Skip("set IDENTITY_IMDS_AVAILABLE to run this test")
 	}
-	opts, stop := initRecording(t)
-	defer stop()
-	cred, err := NewManagedIdentityCredential(&ManagedIdentityCredentialOptions{ClientOptions: opts})
-	if err != nil {
-		t.Fatal(err)
-	}
-	testGetTokenSuccess(t, cred)
-}
 
-func TestManagedIdentityCredential_IMDSClientIDLive(t *testing.T) {
-	switch recording.GetRecordMode() {
-	case recording.LiveMode:
-		t.Skip("this test doesn't run in live mode because it can't pass in CI")
-	case recording.RecordingMode:
-		if liveManagedIdentity.clientID == "" {
-			t.Skip("MANAGED_IDENTITY_CLIENT_ID isn't set")
+	t.Run("client ID", func(t *testing.T) {
+		if recording.GetRecordMode() != recording.PlaybackMode && liveManagedIdentity.clientID == "" {
+			t.Skip("set IDENTITY_VM_USER_ASSIGNED_MI_CLIENT_ID to run this test")
 		}
-	}
-	opts, stop := initRecording(t)
-	defer stop()
-	o := ManagedIdentityCredentialOptions{ClientOptions: opts, ID: ClientID(liveManagedIdentity.clientID)}
-	cred, err := NewManagedIdentityCredential(&o)
-	if err != nil {
-		t.Fatal(err)
-	}
-	testGetTokenSuccess(t, cred)
-}
+		opts, stop := initRecording(t)
+		defer stop()
+		cred, err := NewManagedIdentityCredential(&ManagedIdentityCredentialOptions{
+			ClientOptions: opts, ID: ClientID(liveManagedIdentity.clientID)},
+		)
+		require.NoError(t, err)
+		testGetTokenSuccess(t, cred)
+	})
 
-func TestManagedIdentityCredential_IMDSResourceIDLive(t *testing.T) {
-	switch recording.GetRecordMode() {
-	case recording.LiveMode:
-		t.Skip("this test doesn't run in live mode because it can't pass in CI")
-	case recording.RecordingMode:
-		if liveManagedIdentity.resourceID == "" {
-			t.Skip("MANAGED_IDENTITY_RESOURCE_ID isn't set")
+	t.Run("object ID", func(t *testing.T) {
+		if recording.GetRecordMode() != recording.PlaybackMode && liveManagedIdentity.objectID == "" {
+			t.Skip("set IDENTITY_VM_USER_ASSIGNED_MI_OBJECT_ID to run this test")
 		}
-	}
-	opts, stop := initRecording(t)
-	defer stop()
-	o := ManagedIdentityCredentialOptions{ClientOptions: opts, ID: ResourceID(liveManagedIdentity.resourceID)}
-	cred, err := NewManagedIdentityCredential(&o)
-	if err != nil {
-		t.Fatal(err)
-	}
-	testGetTokenSuccess(t, cred)
+		opts, stop := initRecording(t)
+		defer stop()
+		cred, err := NewManagedIdentityCredential(&ManagedIdentityCredentialOptions{
+			ClientOptions: opts, ID: ObjectID(liveManagedIdentity.objectID)},
+		)
+		require.NoError(t, err)
+		testGetTokenSuccess(t, cred)
+	})
+
+	t.Run("resource ID", func(t *testing.T) {
+		if recording.GetRecordMode() != recording.PlaybackMode && liveManagedIdentity.resourceID == "" {
+			t.Skip("set IDENTITY_VM_USER_ASSIGNED_MI_RESOURCE_ID to run this test")
+		}
+		opts, stop := initRecording(t)
+		defer stop()
+		cred, err := NewManagedIdentityCredential(&ManagedIdentityCredentialOptions{
+			ClientOptions: opts, ID: ResourceID(liveManagedIdentity.resourceID)},
+		)
+		require.NoError(t, err)
+		testGetTokenSuccess(t, cred)
+	})
+
+	t.Run("system assigned", func(t *testing.T) {
+		opts, stop := initRecording(t)
+		defer stop()
+		cred, err := NewManagedIdentityCredential(&ManagedIdentityCredentialOptions{ClientOptions: opts})
+		require.NoError(t, err)
+		testGetTokenSuccess(t, cred)
+	})
 }
 
 func TestManagedIdentityCredential_IMDSRetries(t *testing.T) {
@@ -563,7 +661,6 @@ func TestManagedIdentityCredential_IMDSRetries(t *testing.T) {
 }
 
 func TestManagedIdentityCredential_ServiceFabric(t *testing.T) {
-	resetEnvironmentVarsForTest()
 	expectedSecret := "expected-secret"
 	pred := func(req *http.Request) bool {
 		if secret := req.Header.Get("Secret"); secret != expectedSecret {
@@ -588,3 +685,40 @@ func TestManagedIdentityCredential_ServiceFabric(t *testing.T) {
 	}
 	testGetTokenSuccess(t, cred)
 }
+
+func TestManagedIdentityCredential_UnsupportedID(t *testing.T) {
+	t.Run("Azure Arc", func(t *testing.T) {
+		t.Setenv(identityEndpoint, fakeMIEndpoint)
+		t.Setenv(arcIMDSEndpoint, fakeMIEndpoint)
+		for _, id := range []ManagedIDKind{ClientID(fakeClientID), ObjectID(fakeObjectID), ResourceID(fakeResourceID)} {
+			_, err := NewManagedIdentityCredential(&ManagedIdentityCredentialOptions{ID: id})
+			require.Errorf(t, err, "expected an error for %T", id)
+		}
+	})
+	t.Run("Azure ML", func(t *testing.T) {
+		t.Setenv(msiEndpoint, fakeMIEndpoint)
+		t.Setenv(msiSecret, "...")
+		_, err := NewManagedIdentityCredential(&ManagedIdentityCredentialOptions{ID: ResourceID(fakeResourceID)})
+		require.Error(t, err)
+		_, err = NewManagedIdentityCredential(&ManagedIdentityCredentialOptions{ID: ObjectID(fakeObjectID)})
+		require.Error(t, err)
+		_, err = NewManagedIdentityCredential(&ManagedIdentityCredentialOptions{ID: ClientID(fakeClientID)})
+		require.NoError(t, err)
+	})
+	t.Run("Cloud Shell", func(t *testing.T) {
+		t.Setenv(msiEndpoint, fakeMIEndpoint)
+		for _, id := range []ManagedIDKind{ClientID(fakeClientID), ObjectID(fakeObjectID), ResourceID(fakeResourceID)} {
+			_, err := NewManagedIdentityCredential(&ManagedIdentityCredentialOptions{ID: id})
+			require.Errorf(t, err, "expected an error for %T", id)
+		}
+	})
+	t.Run("Service Fabric", func(t *testing.T) {
+		t.Setenv(identityEndpoint, fakeMIEndpoint)
+		t.Setenv(identityHeader, "...")
+		t.Setenv(identityServerThumbprint, "...")
+		for _, id := range []ManagedIDKind{ClientID(fakeClientID), ObjectID(fakeObjectID), ResourceID(fakeResourceID)} {
+			_, err := NewManagedIdentityCredential(&ManagedIdentityCredentialOptions{ID: id})
+			require.Errorf(t, err, "expected an error for %T", id)
+		}
+	})
+}
