@@ -211,12 +211,18 @@ var lbCreateJSONResponse = `
             "target_port": 80,
             "cdn": {
                 "is_enabled": true
-            }
+            },
+            "region_priorities": {
+                "nyc1": 1,
+                "nyc2": 2
+            },
+            "failover_threshold": 10
         },
         "target_load_balancer_ids": [
             "8268a81c-fcf5-423e-a337-bbfe95817f24",
             "8268a81c-fcf6-423e-a337-bbfe95817f24"
-        ]
+        ],
+		"network": "INTERNAL"
     }
 }
 `
@@ -227,6 +233,7 @@ var lbGetJSONResponse = `
         "id": "37e6be88-01ec-4ec7-9bc6-a514d4719057",
         "name": "example-lb-01",
         "ip": "46.214.185.203",
+        "ipv6": "fd53:616d:6d60::d2c:b001",
         "algorithm": "round_robin",
         "status": "active",
         "created_at": "2016-12-15T14:16:36Z",
@@ -433,6 +440,7 @@ func TestLoadBalancers_Get(t *testing.T) {
 		ID:        "37e6be88-01ec-4ec7-9bc6-a514d4719057",
 		Name:      "example-lb-01",
 		IP:        "46.214.185.203",
+		IPv6:      "fd53:616d:6d60::d2c:b001",
 		Algorithm: "round_robin",
 		Status:    "active",
 		Created:   "2016-12-15T14:16:36Z",
@@ -539,11 +547,14 @@ func TestLoadBalancers_Create(t *testing.T) {
 			{Name: "test-domain-2", IsManaged: true, CertificateID: "test-cert-id-2"},
 		},
 		GLBSettings: &GLBSettings{
-			TargetProtocol: "HTTP",
-			TargetPort:     80,
-			CDN:            &CDNSettings{IsEnabled: true},
+			TargetProtocol:    "HTTP",
+			TargetPort:        80,
+			CDN:               &CDNSettings{IsEnabled: true},
+			RegionPriorities:  map[string]uint32{"nyc1": 1, "nyc2": 2},
+			FailoverThreshold: 10,
 		},
 		TargetLoadBalancerIDs: []string{"8268a81c-fcf5-423e-a337-bbfe95817f24", "8268a81c-fcf6-423e-a337-bbfe95817f24"},
+		Network:               LoadBalancerNetworkTypeInternal,
 	}
 
 	path := "/v2/load_balancers"
@@ -624,11 +635,14 @@ func TestLoadBalancers_Create(t *testing.T) {
 			{Name: "test-domain-2", IsManaged: true, CertificateID: "test-cert-id-2"},
 		},
 		GLBSettings: &GLBSettings{
-			TargetProtocol: "HTTP",
-			TargetPort:     80,
-			CDN:            &CDNSettings{IsEnabled: true},
+			TargetProtocol:    "HTTP",
+			TargetPort:        80,
+			CDN:               &CDNSettings{IsEnabled: true},
+			RegionPriorities:  map[string]uint32{"nyc1": 1, "nyc2": 2},
+			FailoverThreshold: 10,
 		},
 		TargetLoadBalancerIDs: []string{"8268a81c-fcf5-423e-a337-bbfe95817f24", "8268a81c-fcf6-423e-a337-bbfe95817f24"},
+		Network:               LoadBalancerNetworkTypeInternal,
 	}
 
 	disableLetsEncryptDNSRecords := true
@@ -1169,6 +1183,7 @@ func TestLoadBalancers_AsRequest(t *testing.T) {
 			ResponseTimeoutSeconds: 10,
 			HealthyThreshold:       3,
 			UnhealthyThreshold:     3,
+			ProxyProtocol:          PtrTo(true),
 		},
 		StickySessions: &StickySessions{
 			Type:             "cookies",
@@ -1230,6 +1245,7 @@ func TestLoadBalancers_AsRequest(t *testing.T) {
 			ResponseTimeoutSeconds: 10,
 			HealthyThreshold:       3,
 			UnhealthyThreshold:     3,
+			ProxyProtocol:          PtrTo(true),
 		},
 		StickySessions: &StickySessions{
 			Type:             "cookies",
