@@ -153,15 +153,17 @@ func TestStore_BadShard(t *testing.T) {
 			defer require.NoErrorf(t, s.Close(), "closing store with index type: %s", idx)
 
 			sh := tsdb.NewTempShard(idx)
+			shId := sh.ID()
 			err := s.OpenShard(sh.Shard, false)
 			require.NoError(t, err, "opening temp shard")
 			defer require.NoError(t, sh.Close(), "closing temporary shard")
 
-			s.SetShardOpenErrorForTest(sh.ID(), errors.New(errStr))
+			expErr := errors.New(errStr)
+			s.SetShardOpenErrorForTest(sh.ID(), expErr)
 			err2 := s.OpenShard(sh.Shard, false)
 			require.Error(t, err2, "no error opening bad shard")
 			require.True(t, errors.Is(err2, tsdb.ErrPreviousShardFail{}), "exp: ErrPreviousShardFail, got: %v", err2)
-			require.EqualError(t, err2, "opening shard previously failed with: "+errStr)
+			require.EqualError(t, err2, fmt.Errorf("not attempting to open shard %d; opening shard previously failed with: %w", shId, expErr).Error())
 
 			// This should succeed with the force (and because opening an open shard automatically succeeds)
 			require.NoError(t, s.OpenShard(sh.Shard, true), "forced re-opening previously failing shard")
@@ -754,6 +756,82 @@ func TestShards_CreateIterator(t *testing.T) {
 	}
 }
 
+// Test new reader blocking.
+func TestStore_NewReadersBlocked(t *testing.T) {
+	//t.Parallel()
+
+	test := func(index string) {
+		t.Helper()
+		s := MustOpenStore(index)
+		defer s.Close()
+
+		shardInUse := func(shardID uint64) bool {
+			t.Helper()
+			require.NoError(t, s.SetShardNewReadersBlocked(shardID, true))
+			inUse, err := s.ShardInUse(shardID)
+			require.NoError(t, err)
+			require.NoError(t, s.SetShardNewReadersBlocked(shardID, false))
+			return inUse
+		}
+
+		// Create shard #0 with data.
+		s.MustCreateShardWithData("db0", "rp0", 0,
+			`cpu,host=serverA value=1  0`,
+			`cpu,host=serverA value=2 10`,
+			`cpu,host=serverB value=3 20`,
+		)
+
+		// Flush WAL to TSM files.
+		sh0 := s.Shard(0)
+		require.NotNil(t, sh0)
+		sh0.ScheduleFullCompaction()
+
+		// Retrieve shard group.
+		shards := s.ShardGroup([]uint64{0})
+
+		m := &influxql.Measurement{Name: "cpu"}
+		opts := query.IteratorOptions{
+			Expr:       influxql.MustParseExpr(`value`),
+			Dimensions: []string{"host"},
+			Ascending:  true,
+			StartTime:  influxql.MinTime,
+			EndTime:    influxql.MaxTime,
+		}
+
+		// Block new readers, iterator we get will be a faux iterator with no data.
+		require.NoError(t, s.SetShardNewReadersBlocked(0, true))
+		require.False(t, shardInUse(0))
+		itr, err := shards.CreateIterator(context.Background(), m, opts)
+		require.NoError(t, err)
+		require.False(t, shardInUse(0)) // Remember, itr is a faux iterator.
+		fitr, ok := itr.(query.FloatIterator)
+		require.True(t, ok)
+		p, err := fitr.Next()
+		require.NoError(t, err)
+		require.Nil(t, p)
+		require.NoError(t, itr.Close())
+		require.False(t, shardInUse(0))
+		require.NoError(t, s.SetShardNewReadersBlocked(0, false))
+
+		// Create iterator, no blocks present.
+		require.False(t, shardInUse(0))
+		itr, err = shards.CreateIterator(context.Background(), m, opts)
+		require.NoError(t, err)
+		require.True(t, shardInUse(0))
+		fitr, ok = itr.(query.FloatIterator)
+		require.True(t, ok)
+		p, err = fitr.Next()
+		require.NoError(t, err)
+		require.NotNil(t, p)
+		require.NoError(t, itr.Close())
+		require.False(t, shardInUse(0))
+	}
+
+	for _, index := range tsdb.RegisteredIndexes() {
+		t.Run(fmt.Sprintf("TestStore_NewReadersBlocked_%s", index), func(t *testing.T) { test(index) })
+	}
+}
+
 // Ensure the store can backup a shard and another store can restore it.
 func TestStore_BackupRestoreShard(t *testing.T) {
 	test := func(index string) {
